@@ -1,37 +1,58 @@
 """Sphinx autodoc type hints."""
+
 from __future__ import annotations
 
 import ast
+import importlib
 import inspect
 import re
 import sys
 import textwrap
 import types
 from dataclasses import dataclass
-from typing import TYPE_CHECKING, Any, AnyStr, Callable, ForwardRef, NewType, TypeVar, get_type_hints
+from typing import TYPE_CHECKING, Any, AnyStr, ForwardRef, NewType, TypeVar, Union, get_type_hints
 
-from docutils.frontend import OptionParser
-from docutils.parsers.rst import Parser as RstParser
-from docutils.utils import new_document
+from docutils import nodes
+from docutils.frontend import get_default_settings
 from sphinx.ext.autodoc.mock import mock
-from sphinx.util import logging
+from sphinx.parsers import RSTParser
+from sphinx.util import logging, rst
+from sphinx.util.inspect import TypeAliasForwardRef, stringify_signature
 from sphinx.util.inspect import signature as sphinx_signature
-from sphinx.util.inspect import stringify_signature
 
+from ._parser import parse
 from .patches import install_patches
 from .version import __version__
 
 if TYPE_CHECKING:
     from ast import FunctionDef, Module, stmt
+    from collections.abc import Callable
 
     from docutils.nodes import Node
+    from docutils.parsers.rst import states
     from sphinx.application import Sphinx
     from sphinx.config import Config
     from sphinx.environment import BuildEnvironment
     from sphinx.ext.autodoc import Options
 
 _LOGGER = logging.getLogger(__name__)
-_PYDATA_ANNOTATIONS = {"Any", "AnyStr", "Callable", "ClassVar", "Literal", "NoReturn", "Optional", "Tuple", "Union"}
+_PYDATA_ANNOTS_TYPING = {"Any", "AnyStr", "Callable", "ClassVar", "Literal", "NoReturn", "Optional", "Tuple", "Union"}
+_PYDATA_ANNOTS_TYPES = {
+    *("AsyncGeneratorType", "BuiltinFunctionType", "BuiltinMethodType"),
+    *("CellType", "ClassMethodDescriptorType", "CoroutineType"),
+    "EllipsisType",
+    *("FrameType", "FunctionType"),
+    *("GeneratorType", "GetSetDescriptorType"),
+    "LambdaType",
+    *("MemberDescriptorType", "MethodDescriptorType", "MethodType", "MethodWrapperType"),
+    # NoneType is special, but included here for completeness' sake
+    *("NoneType", "NotImplementedType"),
+    "WrapperDescriptorType",
+}
+_PYDATA_ANNOTATIONS = {
+    *(("typing", n) for n in _PYDATA_ANNOTS_TYPING),
+    *(("types", n) for n in _PYDATA_ANNOTS_TYPES),
+}
 
 # types has a bunch of things like ModuleType where ModuleType.__module__ is
 # "builtins" and ModuleType.__name__ is "module", so we have to check for this.
@@ -40,6 +61,11 @@ _TYPES_DICT = {getattr(types, name): name for name in types.__all__}
 _TYPES_DICT[types.FunctionType] = "FunctionType"
 
 
+class MyTypeAliasForwardRef(TypeAliasForwardRef):
+    def __or__(self, value: Any) -> Any:
+        return Union[self, value]  # noqa: UP007
+
+
 def _get_types_type(obj: Any) -> str | None:
     try:
         return _TYPES_DICT.get(obj)
@@ -59,11 +85,11 @@ def get_annotation_module(annotation: Any) -> str:
         return "builtins"
     if _get_types_type(annotation) is not None:
         return "types"
-    is_new_type = sys.version_info >= (3, 10) and isinstance(annotation, NewType)
+    is_new_type = isinstance(annotation, NewType)
     if (
         is_new_type
         or isinstance(annotation, TypeVar)
-        or type(annotation).__name__ in ("ParamSpec", "ParamSpecArgs", "ParamSpecKwargs")
+        or type(annotation).__name__ in {"ParamSpec", "ParamSpecArgs", "ParamSpecKwargs"}
     ):
         return "typing"
     if hasattr(annotation, "__module__"):
@@ -75,8 +101,6 @@ def get_annotation_module(annotation: Any) -> str:
 
 
 def _is_newtype(annotation: Any) -> bool:
-    if sys.version_info < (3, 10):
-        return inspect.isfunction(annotation) and hasattr(annotation, "__supertype__")
     return isinstance(annotation, NewType)
 
 
@@ -103,7 +127,7 @@ def get_annotation_class_name(annotation: Any, module: str) -> str:  # noqa: C90
         return annotation.__qualname__  # type: ignore[no-any-return]
     if getattr(annotation, "_name", None):  # Required for generic aliases on Python 3.7+
         return annotation._name  # type: ignore[no-any-return]  # noqa: SLF001
-    if module in ("typing", "typing_extensions") and isinstance(getattr(annotation, "name", None), str):
+    if module in {"typing", "typing_extensions"} and isinstance(getattr(annotation, "name", None), str):
         # Required for at least Pattern and Match
         return annotation.name  # type: ignore[no-any-return]
 
@@ -136,7 +160,7 @@ def get_annotation_args(annotation: Any, module: str, class_name: str) -> tuple[
             return ()  # This is the original, not parametrized type
 
     # Special cases
-    if class_name in ("Pattern", "Match") and hasattr(annotation, "type_var"):  # Python < 3.7
+    if class_name in {"Pattern", "Match"} and hasattr(annotation, "type_var"):  # Python < 3.7
         return (annotation.type_var,)
     if class_name == "ClassVar" and hasattr(annotation, "__type__"):  # ClassVar on Python < 3.7
         return (annotation.__type__,)
@@ -153,11 +177,11 @@ def get_annotation_args(annotation: Any, module: str, class_name: str) -> tuple[
     return () if len(result) == 1 and result[0] == () else result  # type: ignore[misc]
 
 
-def format_internal_tuple(t: tuple[Any, ...], config: Config) -> str:
-    # An annotation can be a tuple, e.g., for nptyping:
+def format_internal_tuple(t: tuple[Any, ...], config: Config, *, short_literals: bool = False) -> str:
+    # An annotation can be a tuple, e.g., for numpy.typing:
     # In this case, format_annotation receives:
     # This solution should hopefully be general for *any* type that allows tuples in annotations
-    fmt = [format_annotation(a, config) for a in t]
+    fmt = [format_annotation(a, config, short_literals=short_literals) for a in t]
     if len(fmt) == 0:
         return "()"
     if len(fmt) == 1:
@@ -165,12 +189,25 @@ def format_internal_tuple(t: tuple[Any, ...], config: Config) -> str:
     return f"({', '.join(fmt)})"
 
 
-def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PLR0911, PLR0912, PLR0915
+def fixup_module_name(config: Config, module: str) -> str:
+    if getattr(config, "typehints_fixup_module_name", None):
+        module = config.typehints_fixup_module_name(module)
+
+    if module == "typing_extensions":
+        module = "typing"
+
+    if module == "_io":
+        module = "io"
+    return module
+
+
+def format_annotation(annotation: Any, config: Config, *, short_literals: bool = False) -> str:  # noqa: C901, PLR0911, PLR0912, PLR0915, PLR0914
     """
     Format the annotation.
 
     :param annotation:
     :param config:
+    :param short_literals: Render :py:class:`Literals` in PEP 604 style (``|``).
     :return:
     """
     typehints_formatter: Callable[..., str] | None = getattr(config, "typehints_formatter", None)
@@ -190,6 +227,9 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
     if isinstance(annotation, tuple):
         return format_internal_tuple(annotation, config)
 
+    if isinstance(annotation, TypeAliasForwardRef):
+        return annotation.name
+
     try:
         module = get_annotation_module(annotation)
         class_name = get_annotation_class_name(annotation, module)
@@ -197,29 +237,30 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
     except ValueError:
         return str(annotation).strip("'")
 
-    # Redirect all typing_extensions types to the stdlib typing module
-    if module == "typing_extensions":
-        module = "typing"
-
-    if module == "_io":
-        module = "io"
-
+    module = fixup_module_name(config, module)
     full_name = f"{module}.{class_name}" if module != "builtins" else class_name
     fully_qualified: bool = getattr(config, "typehints_fully_qualified", False)
     prefix = "" if fully_qualified or full_name == class_name else "~"
-    role = "data" if module == "typing" and class_name in _PYDATA_ANNOTATIONS else "class"
+    role = "data" if (module, class_name) in _PYDATA_ANNOTATIONS else "class"
     args_format = "\\[{}]"
     formatted_args: str | None = ""
 
+    always_use_bars_union: bool = getattr(config, "always_use_bars_union", True)
+    is_bars_union = full_name == "types.UnionType" or (
+        always_use_bars_union and type(annotation).__qualname__ == "_UnionGenericAlias"
+    )
+    if is_bars_union:
+        full_name = ""
+
     # Some types require special handling
     if full_name == "typing.NewType":
         args_format = f"\\(``{annotation.__name__}``, {{}})"
-        role = "class" if sys.version_info >= (3, 10) else "func"
+        role = "class"
     elif full_name in {"typing.TypeVar", "typing.ParamSpec"}:
         params = {k: getattr(annotation, f"__{k}__") for k in ("bound", "covariant", "contravariant")}
         params = {k: v for k, v in params.items() if v}
         if "bound" in params:
-            params["bound"] = f" {format_annotation(params['bound'], config)}"
+            params["bound"] = f" {format_annotation(params['bound'], config, short_literals=short_literals)}"
         args_format = f"\\(``{annotation.__name__}``{', {}' if args else ''}"
         if params:
             args_format += "".join(f", {k}={v}" for k, v in params.items())
@@ -227,7 +268,7 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
         formatted_args = None if args else args_format
     elif full_name == "typing.Optional":
         args = tuple(x for x in args if x is not type(None))
-    elif full_name in ("typing.Union", "types.UnionType") and type(None) in args:
+    elif full_name in {"typing.Union", "types.UnionType"} and type(None) in args:
         if len(args) == 2:  # noqa: PLR2004
             full_name = "typing.Optional"
             role = "data"
@@ -239,24 +280,27 @@ def format_annotation(annotation: Any, config: Config) -> str:  # noqa: C901, PL
                 role = "data"
                 args_format = f"\\[:py:data:`{prefix}typing.Union`\\[{{}}]]"
                 args = tuple(x for x in args if x is not type(None))
-    elif full_name in ("typing.Callable", "collections.abc.Callable") and args and args[0] is not ...:
-        fmt = [format_annotation(arg, config) for arg in args]
+    elif full_name in {"typing.Callable", "collections.abc.Callable"} and args and args[0] is not ...:
+        fmt = [format_annotation(arg, config, short_literals=short_literals) for arg in args]
         formatted_args = f"\\[\\[{', '.join(fmt[:-1])}], {fmt[-1]}]"
     elif full_name == "typing.Literal":
+        if short_literals:
+            return f"\\{' | '.join(f'``{arg!r}``' for arg in args)}"
         formatted_args = f"\\[{', '.join(f'``{arg!r}``' for arg in args)}]"
-    elif full_name == "types.UnionType":
-        return " | ".join([format_annotation(arg, config) for arg in args])
+    elif is_bars_union:
+        return " | ".join([format_annotation(arg, config, short_literals=short_literals) for arg in args])
 
     if args and not formatted_args:
         try:
             iter(args)
         except TypeError:
-            fmt = [format_annotation(args, config)]
+            fmt = [format_annotation(args, config, short_literals=short_literals)]
         else:
-            fmt = [format_annotation(arg, config) for arg in args]
+            fmt = [format_annotation(arg, config, short_literals=short_literals) for arg in args]
         formatted_args = args_format.format(", ".join(fmt))
 
-    return f":py:{role}:`{prefix}{full_name}`{formatted_args}"
+    escape = "\\ " if formatted_args else ""
+    return f":py:{role}:`{prefix}{full_name}`{escape}{formatted_args}"
 
 
 # reference: https://github.com/pytorch/pytorch/pull/46548/files
@@ -303,7 +347,7 @@ def normalize_source_lines(source_lines: str) -> str:
     return "\n".join(aligned_prefix + aligned_suffix)
 
 
-def process_signature(  # noqa: C901, PLR0913
+def process_signature(  # noqa: C901, PLR0913, PLR0917
     app: Sphinx,
     what: str,
     name: str,
@@ -333,10 +377,25 @@ def process_signature(  # noqa: C901, PLR0913
         return None
 
     obj = inspect.unwrap(obj)
-    sph_signature = sphinx_signature(obj)
+    sph_signature = sphinx_signature(obj, type_aliases=app.config["autodoc_type_aliases"])
+    typehints_formatter: Callable[..., str | None] | None = getattr(app.config, "typehints_formatter", None)
+
+    def _get_formatted_annotation(annotation: TypeVar) -> TypeVar:
+        if typehints_formatter is None:
+            return annotation
+        formatted_name = typehints_formatter(annotation)
+        return annotation if not isinstance(formatted_name, str) else TypeVar(formatted_name)
+
+    if app.config.typehints_use_signature_return:
+        sph_signature = sph_signature.replace(
+            return_annotation=_get_formatted_annotation(sph_signature.return_annotation)
+        )
 
     if app.config.typehints_use_signature:
-        parameters = list(sph_signature.parameters.values())
+        parameters = [
+            param.replace(annotation=_get_formatted_annotation(param.annotation))
+            for param in sph_signature.parameters.values()
+        ]
     else:
         parameters = [param.replace(annotation=inspect.Parameter.empty) for param in sph_signature.parameters.values()]
 
@@ -358,14 +417,20 @@ def process_signature(  # noqa: C901, PLR0913
                 # when method starts with double underscore Python applies mangling -> prepend the class name
                 method_name = f"_{obj.__qualname__.split('.')[-2]}{method_name}"
             method_object = outer.__dict__[method_name] if outer else obj
-            if not isinstance(method_object, (classmethod, staticmethod)):
+            if not isinstance(method_object, classmethod | staticmethod):
                 start = 1
 
     sph_signature = sph_signature.replace(parameters=parameters[start:])
-    if not app.config.typehints_use_signature_return:
-        sph_signature = sph_signature.replace(return_annotation=inspect.Signature.empty)
-
-    return stringify_signature(sph_signature).replace("\\", "\\\\"), None
+    show_return_annotation = app.config.typehints_use_signature_return
+    unqualified_typehints = not getattr(app.config, "typehints_fully_qualified", False)
+    return (
+        stringify_signature(
+            sph_signature,
+            show_return_annotation=show_return_annotation,
+            unqualified_typehints=unqualified_typehints,
+        ).replace("\\", "\\\\"),
+        None,
+    )
 
 
 def _is_dataclass(name: str, what: str, qualname: str) -> bool:
@@ -376,18 +441,19 @@ def _is_dataclass(name: str, what: str, qualname: str) -> bool:
 
 
 def _future_annotations_imported(obj: Any) -> bool:
-    _annotations = getattr(inspect.getmodule(obj), "annotations", None)
-    if _annotations is None:
+    annotations_ = getattr(inspect.getmodule(obj), "annotations", None)
+    if annotations_ is None:
         return False
 
     # Make sure that annotations is imported from __future__ - defined in cpython/Lib/__future__.py
     # annotations become strings at runtime
-    future_annotations = 0x100000 if sys.version_info[0:2] == (3, 7) else 0x1000000
-    return bool(_annotations.compiler_flag == future_annotations)
+    return bool(annotations_.compiler_flag == 0x1000000)  # pragma: no cover # noqa: PLR2004
 
 
-def get_all_type_hints(autodoc_mock_imports: list[str], obj: Any, name: str) -> dict[str, Any]:
-    result = _get_type_hint(autodoc_mock_imports, name, obj)
+def get_all_type_hints(
+    autodoc_mock_imports: list[str], obj: Any, name: str, localns: dict[Any, MyTypeAliasForwardRef]
+) -> dict[str, Any]:
+    result = _get_type_hint(autodoc_mock_imports, name, obj, localns)
     if not result:
         result = backfill_type_hints(obj, name)
         try:
@@ -395,7 +461,7 @@ def get_all_type_hints(autodoc_mock_imports: list[str], obj: Any, name: str) ->
         except (AttributeError, TypeError):
             pass
         else:
-            result = _get_type_hint(autodoc_mock_imports, name, obj)
+            result = _get_type_hint(autodoc_mock_imports, name, obj, localns)
     return result
 
 
@@ -404,38 +470,64 @@ _TYPE_GUARD_IMPORTS_RESOLVED = set()
 _TYPE_GUARD_IMPORTS_RESOLVED_GLOBALS_ID = set()
 
 
-def _resolve_type_guarded_imports(autodoc_mock_imports: list[str], obj: Any) -> None:  # noqa: C901
-    if hasattr(obj, "__module__") and obj.__module__ in _TYPE_GUARD_IMPORTS_RESOLVED:
-        return  # already processed module
-    if not hasattr(obj, "__globals__"):  # classes with __slots__ do not have this
-        return  # if lacks globals nothing we can do
-    if id(obj.__globals__) in _TYPE_GUARD_IMPORTS_RESOLVED_GLOBALS_ID:
-        return  # already processed object
-    _TYPE_GUARD_IMPORTS_RESOLVED.add(obj.__module__)
-    if obj.__module__ not in sys.builtin_module_names:
-        if hasattr(obj, "__globals__"):
-            _TYPE_GUARD_IMPORTS_RESOLVED_GLOBALS_ID.add(id(obj.__globals__))
+def _should_skip_guarded_import_resolution(obj: Any) -> bool:
+    if isinstance(obj, types.ModuleType):
+        return False  # Don't skip modules
 
-        module = inspect.getmodule(obj)
-        if module:
+    if not hasattr(obj, "__globals__"):
+        return True  # Skip objects without __globals__
+
+    if hasattr(obj, "__module__"):
+        return obj.__module__ in _TYPE_GUARD_IMPORTS_RESOLVED or obj.__module__ in sys.builtin_module_names
+
+    return id(obj.__globals__) in _TYPE_GUARD_IMPORTS_RESOLVED_GLOBALS_ID
+
+
+def _execute_guarded_code(autodoc_mock_imports: list[str], obj: Any, module_code: str) -> None:
+    for _, part in _TYPE_GUARD_IMPORT_RE.findall(module_code):
+        guarded_code = textwrap.dedent(part)
+        try:
             try:
-                module_code = inspect.getsource(module)
-            except (TypeError, OSError):
-                ...  # no source code => no type guards
-            else:
-                for _, part in _TYPE_GUARD_IMPORT_RE.findall(module_code):
-                    guarded_code = textwrap.dedent(part)
-                    try:
-                        with mock(autodoc_mock_imports):
-                            exec(guarded_code, obj.__globals__)  # noqa: S102
-                    except Exception as exc:  # noqa: BLE001, PERF203
-                        _LOGGER.warning(f"Failed guarded type import with {exc!r}")
+                with mock(autodoc_mock_imports):
+                    exec(guarded_code, getattr(obj, "__globals__", obj.__dict__))  # noqa: S102
+            except ImportError as exc:
+                # ImportError might have occurred because the module has guarded code as well,
+                # so we recurse on the module.
+                if exc.name:
+                    _resolve_type_guarded_imports(autodoc_mock_imports, importlib.import_module(exc.name))
+
+                    # Retry the guarded code and see if it works now after resolving all nested type guards.
+                    with mock(autodoc_mock_imports):
+                        exec(guarded_code, getattr(obj, "__globals__", obj.__dict__))  # noqa: S102
+        except Exception as exc:  # noqa: BLE001
+            _LOGGER.warning("Failed guarded type import with %r", exc)
 
 
-def _get_type_hint(autodoc_mock_imports: list[str], name: str, obj: Any) -> dict[str, Any]:
+def _resolve_type_guarded_imports(autodoc_mock_imports: list[str], obj: Any) -> None:
+    if _should_skip_guarded_import_resolution(obj):
+        return
+
+    if hasattr(obj, "__globals__"):
+        _TYPE_GUARD_IMPORTS_RESOLVED_GLOBALS_ID.add(id(obj.__globals__))
+
+    module = inspect.getmodule(obj)
+
+    if module:
+        try:
+            module_code = inspect.getsource(module)
+        except (TypeError, OSError):
+            ...  # no source code => no type guards
+        else:
+            _TYPE_GUARD_IMPORTS_RESOLVED.add(module.__name__)
+            _execute_guarded_code(autodoc_mock_imports, obj, module_code)
+
+
+def _get_type_hint(
+    autodoc_mock_imports: list[str], name: str, obj: Any, localns: dict[Any, MyTypeAliasForwardRef]
+) -> dict[str, Any]:
     _resolve_type_guarded_imports(autodoc_mock_imports, obj)
     try:
-        result = get_type_hints(obj)
+        result = get_type_hints(obj, None, localns)
     except (AttributeError, TypeError, RecursionError) as exc:
         # TypeError - slot wrapper, PEP-563 when part of new syntax not supported
         # RecursionError - some recursive type definitions https://github.com/python/typing/issues/574
@@ -449,7 +541,7 @@ def _get_type_hint(autodoc_mock_imports: list[str], name: str, obj: Any) -> dict
     return result
 
 
-def backfill_type_hints(obj: Any, name: str) -> dict[str, Any]:  # noqa: C901, PLR0911, PLR0912
+def backfill_type_hints(obj: Any, name: str) -> dict[str, Any]:  # noqa: C901, PLR0911
     """
     Backfill type hints.
 
@@ -498,7 +590,7 @@ def backfill_type_hints(obj: Any, name: str) -> dict[str, Any]:  # noqa: C901, P
     comment_args = split_type_comment_args(comment_args_str)
     is_inline = len(comment_args) == 1 and comment_args[0] == "..."
     if not is_inline:
-        if args and args[0].arg in ("self", "cls") and len(comment_args) != len(args):
+        if args and args[0].arg in {"self", "cls"} and len(comment_args) != len(args):
             comment_args.insert(0, None)  # self/cls may be omitted in type comments, insert blank
 
         if len(args) != len(comment_args):
@@ -547,9 +639,9 @@ def split_type_comment_args(comment: str) -> list[str | None]:
 
     brackets, start_arg_at, at = 0, 0, 0
     for at, char in enumerate(comment):
-        if char in ("[", "("):
+        if char in {"[", "("}:
             brackets += 1
-        elif char in ("]", ")"):
+        elif char in {"]", ")"}:
             brackets -= 1
         elif char == "," and brackets == 0:
             add(comment[start_arg_at:at])
@@ -573,7 +665,7 @@ def format_default(app: Sphinx, default: Any, is_annotated: bool) -> str | None:
     return f"default: ``{formatted}``"
 
 
-def process_docstring(  # noqa: PLR0913
+def process_docstring(  # noqa: PLR0913, PLR0917
     app: Sphinx,
     what: str,
     name: str,
@@ -600,11 +692,13 @@ def process_docstring(  # noqa: PLR0913
     obj = inspect.unwrap(obj)
 
     try:
-        signature = sphinx_signature(obj)
+        signature = sphinx_signature(obj, type_aliases=app.config["autodoc_type_aliases"])
     except (ValueError, TypeError):
         signature = None
-    type_hints = get_all_type_hints(app.config.autodoc_mock_imports, obj, name)
-    app.config._annotation_globals = getattr(obj, "__globals__", {})  # type: ignore[attr-defined]  # noqa: SLF001
+
+    localns = {key: MyTypeAliasForwardRef(value) for key, value in app.config["autodoc_type_aliases"].items()}
+    type_hints = get_all_type_hints(app.config.autodoc_mock_imports, obj, name, localns)
+    app.config._annotation_globals = getattr(obj, "__globals__", {})  # noqa: SLF001
     try:
         _inject_types_to_docstring(type_hints, signature, original_obj, app, what, name, lines)
     finally:
@@ -652,7 +746,7 @@ def _line_is_param_line_for_arg(line: str, arg_name: str) -> bool:
     return any(doc_name == prefix + arg_name for prefix in ("", "\\*", "\\**", "\\*\\*"))
 
 
-def _inject_types_to_docstring(  # noqa: PLR0913
+def _inject_types_to_docstring(  # noqa: PLR0913, PLR0917
     type_hints: dict[str, Any],
     signature: inspect.Signature | None,
     original_obj: Any,
@@ -667,14 +761,14 @@ def _inject_types_to_docstring(  # noqa: PLR0913
         _inject_rtype(type_hints, original_obj, app, what, name, lines)
 
 
-def _inject_signature(  # noqa: C901
+def _inject_signature(
     type_hints: dict[str, Any],
     signature: inspect.Signature,
     app: Sphinx,
     lines: list[str],
 ) -> None:
     for arg_name in signature.parameters:
-        annotation = type_hints.get(arg_name, None)
+        annotation = type_hints.get(arg_name)
 
         default = signature.parameters[arg_name].default
 
@@ -699,20 +793,42 @@ def _inject_signature(  # noqa: C901
             if annotation is None:
                 type_annotation = f":type {arg_name}: "
             else:
-                formatted_annotation = format_annotation(annotation, app.config)
+                short_literals = app.config.python_display_short_literal_types
+                formatted_annotation = add_type_css_class(
+                    format_annotation(annotation, app.config, short_literals=short_literals)
+                )
                 type_annotation = f":type {arg_name}: {formatted_annotation}"
 
             if app.config.typehints_defaults:
                 formatted_default = format_default(app, default, annotation is not None)
                 if formatted_default:
-                    if app.config.typehints_defaults.endswith("after"):
-                        lines[insert_index] += formatted_default
-                    else:  # add to last param doc line
-                        type_annotation += formatted_default
+                    type_annotation = _append_default(app, lines, insert_index, type_annotation, formatted_default)
 
             lines.insert(insert_index, type_annotation)
 
 
+def _append_default(
+    app: Sphinx, lines: list[str], insert_index: int, type_annotation: str, formatted_default: str
+) -> str:
+    if app.config.typehints_defaults.endswith("after"):
+        # advance the index to the end of the :param: paragraphs
+        # (terminated by a line with no indentation)
+        # append default to the last nonempty line
+        nlines = len(lines)
+        next_index = insert_index + 1
+        append_index = insert_index  # last nonempty line
+        while next_index < nlines and (not lines[next_index] or lines[next_index].startswith(" ")):
+            if lines[next_index]:
+                append_index = next_index
+            next_index += 1
+        lines[append_index] += formatted_default
+
+    else:  # add to last param doc line
+        type_annotation += formatted_default
+
+    return type_annotation
+
+
 @dataclass
 class InsertIndexInfo:
     insert_index: int
@@ -736,6 +852,9 @@ def node_line_no(node: Node) -> int | None:
     docutils rst parser source code. An example where the node doesn't have a line number but the first child does is
     all `definition_list` nodes. It seems like bullet_list and option_list get line numbers, but enum_list also doesn't.
     """
+    if node is None:
+        return None
+
     while node.line is None and node.children:
         node = node.children[0]
     return node.line
@@ -758,10 +877,9 @@ def get_insert_index(app: Sphinx, lines: list[str]) -> InsertIndexInfo | None:
 
     # 3. Insert after the parameters.
     # To find the parameters, parse as a docutils tree.
-    settings = OptionParser(components=(RstParser,)).get_default_values()
+    settings = get_default_settings(RSTParser)
     settings.env = app.env
-    doc = new_document("", settings=settings)
-    RstParser().parse("\n".join(lines), doc)
+    doc = parse("\n".join(lines), settings)
 
     # Find a top level child which is a field_list that contains a field whose
     # name starts with one of the PARAM_SYNONYMS. This is the parameter list. We
@@ -779,22 +897,24 @@ def get_insert_index(app: Sphinx, lines: list[str]) -> InsertIndexInfo | None:
         # at end. (I don't know of any input where this happens.)
         next_sibling = child.next_node(descend=False, siblings=True)
         line_no = node_line_no(next_sibling) if next_sibling else None
-        at = line_no - 2 if line_no else len(lines)
+        at = max(line_no - 2, 0) if line_no else len(lines)
         return InsertIndexInfo(insert_index=at, found_param=True)
 
     # 4. Insert before examples
     for child in doc.children:
-        if tag_name(child) in ["literal_block", "paragraph", "field_list"]:
+        if tag_name(child) in {"literal_block", "paragraph", "field_list"}:
             continue
         line_no = node_line_no(child)
-        at = line_no - 2 if line_no else len(lines)
+        at = max(line_no - 2, 0) if line_no else len(lines)
+        if lines[at - 1]:  # skip if something on this line
+            break
         return InsertIndexInfo(insert_index=at, found_directive=True)
 
     # 5. Otherwise, insert at end
     return InsertIndexInfo(insert_index=len(lines))
 
 
-def _inject_rtype(  # noqa: PLR0913
+def _inject_rtype(  # noqa: C901, PLR0913, PLR0917
     type_hints: dict[str, Any],
     original_obj: Any,
     app: Sphinx,
@@ -808,6 +928,8 @@ def _inject_rtype(  # noqa: PLR0913
         return
     if not app.config.typehints_document_rtype:
         return
+    if not app.config.typehints_document_rtype_none and type_hints["return"] is types.NoneType:
+        return
 
     r = get_insert_index(app, lines)
     if r is None:
@@ -818,7 +940,10 @@ def _inject_rtype(  # noqa: PLR0913
     if not app.config.typehints_use_rtype and r.found_return and " -- " in lines[insert_index]:
         return
 
-    formatted_annotation = format_annotation(type_hints["return"], app.config)
+    short_literals = app.config.python_display_short_literal_types
+    formatted_annotation = add_type_css_class(
+        format_annotation(type_hints["return"], app.config, short_literals=short_literals)
+    )
 
     if r.found_param and insert_index < len(lines) and lines[insert_index].strip():
         insert_index -= 1
@@ -834,7 +959,7 @@ def _inject_rtype(  # noqa: PLR0913
             lines.insert(insert_index + 1, "")
     else:
         line = lines[insert_index]
-        lines[insert_index] = f":return: {formatted_annotation} --{line[line.find(' '):]}"
+        lines[insert_index] = f":return: {formatted_annotation} --{line[line.find(' ') :]}"
 
 
 def validate_config(app: Sphinx, env: BuildEnvironment, docnames: list[str]) -> None:  # noqa: ARG001
@@ -849,16 +974,57 @@ def validate_config(app: Sphinx, env: BuildEnvironment, docnames: list[str]) ->
         raise ValueError(msg)
 
 
+def unescape(escaped: str) -> str:
+    # For some reason the string we get has a bunch of null bytes in it??
+    # Remove them...
+    escaped = escaped.replace("\x00", "")
+    # For some reason the extra slash before spaces gets lost between the .rst
+    # source and when this directive is called. So don't replace "\<space>" =>
+    # "<space>"
+    return re.sub(r"\\([^ ])", r"\1", escaped)
+
+
+def add_type_css_class(type_rst: str) -> str:
+    return f":sphinx_autodoc_typehints_type:`{rst.escape(type_rst)}`"
+
+
+def sphinx_autodoc_typehints_type_role(
+    _role: str,
+    _rawtext: str,
+    text: str,
+    _lineno: int,
+    inliner: states.Inliner,
+    _options: dict[str, Any] | None = None,
+    _content: list[str] | None = None,
+) -> tuple[list[Node], list[Node]]:
+    """
+    Add css tag around rendered type.
+
+    The body should be escaped rst. This renders its body as rst and wraps the
+    result in <span class="sphinx_autodoc_typehints-type"> </span>
+    """
+    unescaped = unescape(text)
+    doc = parse(unescaped, inliner.document.settings)
+    n = nodes.inline(text)
+    n["classes"].append("sphinx_autodoc_typehints-type")
+    n += doc.children[0].children
+    return [n], []
+
+
 def setup(app: Sphinx) -> dict[str, bool]:
     app.add_config_value("always_document_param_types", False, "html")  # noqa: FBT003
     app.add_config_value("typehints_fully_qualified", False, "env")  # noqa: FBT003
     app.add_config_value("typehints_document_rtype", True, "env")  # noqa: FBT003
+    app.add_config_value("typehints_document_rtype_none", True, "env")  # noqa: FBT003
     app.add_config_value("typehints_use_rtype", True, "env")  # noqa: FBT003
     app.add_config_value("typehints_defaults", None, "env")
     app.add_config_value("simplify_optional_unions", True, "env")  # noqa: FBT003
+    app.add_config_value("always_use_bars_union", False, "env")  # noqa: FBT003
     app.add_config_value("typehints_formatter", None, "env")
     app.add_config_value("typehints_use_signature", False, "env")  # noqa: FBT003
     app.add_config_value("typehints_use_signature_return", False, "env")  # noqa: FBT003
+    app.add_config_value("typehints_fixup_module_name", None, "env")
+    app.add_role("sphinx_autodoc_typehints_type", sphinx_autodoc_typehints_type_role)
     app.connect("env-before-read-docs", validate_config)  # config may be changed after “config-inited” event
     app.connect("autodoc-process-signature", process_signature)
     app.connect("autodoc-process-docstring", process_docstring)
@@ -868,6 +1034,7 @@ def setup(app: Sphinx) -> dict[str, bool]:
 
 __all__ = [
     "__version__",
+    "backfill_type_hints",
     "format_annotation",
     "get_annotation_args",
     "get_annotation_class_name",
@@ -875,5 +1042,4 @@ __all__ = [
     "normalize_source_lines",
     "process_docstring",
     "process_signature",
-    "backfill_type_hints",
 ]
