@@ -1,16 +1,19 @@
 """Module demonstrating imports that are type guarded"""
+
 from __future__ import annotations
 
 import typing
-from builtins import ValueError  # handle does not have __module__
+from builtins import ValueError  # handle does not have __module__  # noqa: A004
 from functools import cmp_to_key  # has __module__ but cannot get module as is builtin
 from typing import TYPE_CHECKING
 
 from demo_typing_guard_dummy import AnotherClass
 
 if TYPE_CHECKING:
+    from collections.abc import Sequence
     from decimal import Decimal
-    from typing import Sequence
+
+    from demo_typing_guard_dummy import Literal  # guarded by another `if TYPE_CHECKING` in demo_typing_guard_dummy
 
 
 if typing.TYPE_CHECKING:
@@ -52,10 +55,13 @@ class SomeClass:
             """
 
 
+def func(_x: Literal) -> None: ...
+
+
 __all__ = [
-    "a",
-    "ValueError",
-    "cmp_to_key",
-    "SomeClass",
     "AnotherClass",
+    "SomeClass",
+    "ValueError",
+    "a",
+    "cmp_to_key",
 ]
