@@ -1319,6 +1319,19 @@ def typehints_use_signature(a: AsyncGenerator) -> AsyncGenerator:
     return a
 
 
+@expected(
+    """
+    mod.typehints_no_rtype_none()
+
+       Do something.
+
+    """,
+    typehints_document_rtype_none=False,
+)
+def typehints_no_rtype_none() -> None:
+    """Do something."""
+
+
 prolog = """
 .. |test_node_start| replace:: {test_node_start}
 """.format(test_node_start="test_start")
@@ -1397,15 +1410,19 @@ def docstring_with_multiline_note_after_params_epilog_replace(param: int) -> Non
     """
     mod.docstring_with_see_also()
 
-       Return type:
-          "str"
+       Test
 
        See also: more info at <https://example.com>`_.
 
+       Return type:
+          "str"
+
     """
 )
 def docstring_with_see_also() -> str:
     """
+    Test
+
     .. seealso:: more info at <https://example.com>`_.
     """
     return ""
